@@ -3,6 +3,7 @@
 #define NTL_vector__H
 
 #include <NTL/tools.h>
+#include <new>
 
 struct _ntl_VectorHeader {
    long length;
@@ -16,37 +17,18 @@ union _ntl_AlignedVectorHeader {
    double x1;
    long x2;
    char *x3;
+   long double x4;
 };
 
 #define NTL_VECTOR_HEADER_SIZE (sizeof(_ntl_AlignedVectorHeader))
 
-#define NTL_VEC_HEAD(p) (& (((_ntl_AlignedVectorHeader *) p)[-1].h))
+#define NTL_VEC_HEAD(p) (& (((_ntl_AlignedVectorHeader *) (p.rep))[-1].h))
 
-struct _ntl_vector_placement {
-   void *p;
-};
-
-inline _ntl_vector_placement _ntl_vector_placement_fn(void *p)
-{
-   _ntl_vector_placement x;
-   x.p = p;
-   return x;
-}
-
-inline void *operator new(NTL_SNS size_t, _ntl_vector_placement x) { return x.p; }
-
-// All of this monkey business is to avoid possible clashes with
-// a "placement new" operator which may or may not be defined
-// in a standard header file....why wasn't this just built
-// into the language to begin with?
 
 #ifndef NTL_RANGE_CHECK
-#define NTL_RANGE_CHECK_CODE 
-#define NTL_RANGE_CHECK_CODE1(i) 
+#define NTL_RANGE_CHECK_CODE(i) 
 #else
-#define NTL_RANGE_CHECK_CODE if (l__i < 0 || !_vec__rep || l__i >= NTL_VEC_HEAD(_vec__rep)->length) RangeError(l__i);
-
-#define NTL_RANGE_CHECK_CODE1(i) if ((i) < 0 || !_vec__rep || (i) >= NTL_VEC_HEAD(_vec__rep)->length) RangeError(i);
+#define NTL_RANGE_CHECK_CODE(i) if ((i) < 0 || !_vec__rep || (i) >= NTL_VEC_HEAD(_vec__rep)->length) LogicError("index out of range in Vec");
 
 #endif
 
@@ -71,54 +53,123 @@ inline void *operator new(NTL_SNS size_t, _ntl_vector_placement x) { return x.p;
 
 NTL_OPEN_NNS
 
-
-template<class T>
-void BlockConstruct(T* p, long n)  
-{  
-   for (long i = 0; i < n; i++)  
-      (void) new(_ntl_vector_placement_fn(&p[i])) T;  
-}  
-
-template<class T>
-void BlockConstructFromVec(T* p, long n, const T* q)  
-{  
-   for (long i = 0; i < n; i++)  
-      (void) new(_ntl_vector_placement_fn(&p[i])) T(q[i]);  
-}  
-
-template<class T>
-void BlockConstructFromObj(T* p, long n, const T& q)  
-{  
-   for (long i = 0; i < n; i++)  
-      (void) new(_ntl_vector_placement_fn(&p[i])) T(q);  
-}  
-
   
+
+
+
 template<class T>
 void BlockDestroy(T* p, long n)  
 {  
    for (long i = 0; i < n; i++)  
       p[i].~T();  
+
+   // NOTE: this routine is only invoked through a Vec destructor
+   // or a scope guard destructor, both of which are noexcept destructors.
+   // therefore, if ~T() should throw, the program will terminate
 }
 
 
+template<class T>
+void BlockConstruct(T* p, long n)  
+{  
+   long i;
+
+   NTL_SCOPE(guard) { BlockDestroy(p, i); };
+
+   for (i = 0; i < n; i++)  
+      (void) new(&p[i]) T;  
+
+   guard.relax();
+
+   // NOTE: we invoke T rather than T(), which would ensure
+   // POD types get zeroed out, but only in compilers that
+   // comply with C++03, which does not include MS compilers.
+   // So we just use T, which is less expensive, and it is better
+   // not to assume POD types get initialized.
+   
+}  
+
+template<class T>
+void BlockConstructFromVec(T* p, long n, const T* q)  
+{  
+   long i;
+
+   NTL_SCOPE(guard) { BlockDestroy(p, i); };
+
+   for (i = 0; i < n; i++)  
+      (void) new(&p[i]) T(q[i]);  
+
+   guard.relax();
+}  
+
+template<class T>
+void BlockConstructFromObj(T* p, long n, const T& q)  
+{  
+   long i;
+
+   NTL_SCOPE(guard) { BlockDestroy(p, i); };
+
+   for (i = 0; i < n; i++)  
+      (void) new(&p[i]) T(q);  
+
+   guard.relax();
+}  
+
+// experimental stuff to implement Vec<T> for non-relocatable T
+// Right now, it is not used anywhere
+
+
+struct RelocatableTypeTag { };
+struct NotRelocatableTypeTag { };
+
+template<class T>
+RelocatableTypeTag IsRelocatableType(T*) { return RelocatableTypeTag(); }
+
+
 template<class T>
 class Vec {  
 public:  
 
-   T *_vec__rep;  
+   static bool relocatable(RelocatableTypeTag) { return true; }
+   static bool relocatable(NotRelocatableTypeTag) { return false; }
+   static bool relocatable() { return relocatable(IsRelocatableType((T*)0)); }
+
+   class _vec_deleter {
+   public:
+      static void apply(T* p) { 
+         if (p)  {
+            NTL_SNS free(((char *) p) - sizeof(_ntl_AlignedVectorHeader));
+         }
+      }
+   };
+
+   WrappedPtr<T, _vec_deleter> _vec__rep;  
   
-   void RangeError(long i) const;  
   
-   Vec() : _vec__rep(0) { }  
-   Vec(INIT_SIZE_TYPE, long n) : _vec__rep(0) { SetLength(n); }  
-   Vec(const Vec<T>& a) : _vec__rep(0) { *this = a; }     
+   Vec() { }  
+
+   Vec(INIT_SIZE_TYPE, long n) { SetLength(n); }  
+   Vec(INIT_SIZE_TYPE, long n, const T& a) { SetLength(n, a); }  
+
+   // the following copy constructor does not rely on
+   // the assignment operator
+   Vec(const Vec<T>& a)  
+   {  
+      long src_len = a.length();
+      const T *src = a.elts();
+      AllocateTo(src_len);
+      Init(src_len, src);
+      AdjustLength(src_len);
+   }     
+
    Vec<T>& operator=(const Vec<T>& a);  
+
    ~Vec();  
    void kill(); 
   
    void SetMaxLength(long n); 
    void FixLength(long n); 
+   void FixAtCurrentLength();
    void QuickSetLength(long n) { NTL_VEC_HEAD(_vec__rep)->length = n; } 
 
    void SetLength(long n) {
@@ -137,6 +188,16 @@ public:
          DoSetLength(n, a);
    }
 
+   template<class F>
+   void SetLengthAndApply(long n, F f) {
+      if (_vec__rep && !NTL_VEC_HEAD(_vec__rep)->fixed &&
+          n >= 0 && n <= NTL_VEC_HEAD(_vec__rep)->init)
+         NTL_VEC_HEAD(_vec__rep)->length = n;
+      else 
+         DoSetLengthAndApply(n, f);
+   }
+
+
 
    long length() const 
    { return (!_vec__rep) ?  0 : NTL_VEC_HEAD(_vec__rep)->length; }  
@@ -152,13 +213,13 @@ public:
   
    T& operator[](long i)   
    {  
-      NTL_RANGE_CHECK_CODE1(i)  
+      NTL_RANGE_CHECK_CODE(i)  
       return _vec__rep[i];  
    }  
   
    const T& operator[](long i) const 
    {  
-      NTL_RANGE_CHECK_CODE1(i)  
+      NTL_RANGE_CHECK_CODE(i)  
       return _vec__rep[i];  
    }  
   
@@ -179,9 +240,8 @@ public:
    const T* elts() const { return _vec__rep; }  
    T* elts() { return _vec__rep; }  
          
- 
    Vec(Vec<T>& x, INIT_TRANS_TYPE) 
-   { _vec__rep = x._vec__rep; x._vec__rep = 0; } 
+   { _vec__rep.swap(x._vec__rep); }
 
    long position(const T& a) const;  
    long position1(const T& a) const;  
@@ -230,29 +290,49 @@ public:
 
    T& at(long i) {
       if ((i) < 0 || !_vec__rep || (i) >= NTL_VEC_HEAD(_vec__rep)->length)  
-         RangeError(i);
+         LogicError("index out of range in Vec");
       return _vec__rep[i];  
    }
 
    const T& at(long i) const {
       if ((i) < 0 || !_vec__rep || (i) >= NTL_VEC_HEAD(_vec__rep)->length)  
-         RangeError(i);
+         LogicError("index out of range in Vec");
       return _vec__rep[i];  
    }
 
+   class Watcher {
+   public:
+      Vec& watched;
+      explicit
+      Watcher(Vec& _watched) : watched(_watched) {}
+
+      ~Watcher() 
+      { 
+         if (watched.MaxLength() > NTL_RELEASE_THRESH) watched.kill();
+      }
+   };
 
 private:
    void DoSetLength(long n);
    void DoSetLength(long n, const T& a);
 
+   template<class F>
+   void DoSetLengthAndApply(long n, F& f);
+
    void AdjustLength(long n) { if (_vec__rep) NTL_VEC_HEAD(_vec__rep)->length = n; }
    void AdjustAlloc(long n) { if (_vec__rep) NTL_VEC_HEAD(_vec__rep)->alloc = n; }
    void AdjustMaxLength(long n) { if (_vec__rep) NTL_VEC_HEAD(_vec__rep)->init = n; }
 
-   void AllocateTo(long n); // reserves space for n items, also checking
+   void ReAllocate(long n, RelocatableTypeTag);
+   void ReAllocate(long n, NotRelocatableTypeTag);
+
+   void AllocateTo(long n); // reserves space for n items
    void Init(long n); // make sure first n entries are initialized
    void Init(long n, const T* src); // same, but use src
    void Init(long n, const T& src); // same, but use src
+
+   template<class F>
+   void InitAndApply(long n, F& f);
 };  
  
 
@@ -273,7 +353,7 @@ long Vec<T>::position(const T& a) const
        _vec__rep + res != &a) return -1;  
    
    if (res >= num_init)  
-       Error("position: reference to uninitialized object"); 
+       LogicError("position: reference to uninitialized object"); 
    return res;  
 }  
   
@@ -305,7 +385,7 @@ long Vec<T>::position(const T& a) const
    while (res < num_alloc && _vec__rep + res != &a)  res++;  
    if (res >= num_alloc) return -1;  
    if (res >= num_init)  
-       Error("position: reference to uninitialized object"); 
+       LogicError("position: reference to uninitialized object"); 
    return res;  
 }  
  
@@ -324,6 +404,41 @@ long Vec<T>::position1(const T& a) const
 
 #endif
 
+template<class T>
+void Vec<T>::ReAllocate(long m, RelocatableTypeTag)   
+{
+   char *p = ((char *) _vec__rep.rep) - sizeof(_ntl_AlignedVectorHeader); 
+   p = (char *) NTL_SNS_REALLOC(p, m, sizeof(T), sizeof(_ntl_AlignedVectorHeader)); 
+   if (!p) {  
+      MemoryError();  
+   }  
+   _vec__rep = (T *) (p + sizeof(_ntl_AlignedVectorHeader)); 
+   NTL_VEC_HEAD(_vec__rep)->alloc = m;  
+}
+
+template<class T>
+void Vec<T>::ReAllocate(long m, NotRelocatableTypeTag)   
+{
+   Vec<T> tmp;
+   long src_len = length();
+   long src_init = MaxLength();
+   const T *src = elts();
+
+   char *p = (char *) NTL_SNS_MALLOC(m, sizeof(T), sizeof(_ntl_AlignedVectorHeader)); 
+   if (!p) {  
+      MemoryError();  
+   }  
+   tmp._vec__rep = (T *) (p + sizeof(_ntl_AlignedVectorHeader)); 
+
+   NTL_VEC_HEAD(tmp._vec__rep)->length = 0;  
+   NTL_VEC_HEAD(tmp._vec__rep)->alloc = m;  
+   NTL_VEC_HEAD(tmp._vec__rep)->init = 0;  
+   NTL_VEC_HEAD(tmp._vec__rep)->fixed = 0;  
+
+   tmp.Init(src_init, src);    // copy all initialized elements
+   tmp.AdjustLength(src_len);
+   tmp.swap(*this);
+}
  
 template<class T>
 void Vec<T>::AllocateTo(long n)   
@@ -331,16 +446,16 @@ void Vec<T>::AllocateTo(long n)
    long m;  
   
    if (n < 0) {  
-      Error("negative length in vector::SetLength");  
+      LogicError("negative length in vector::SetLength");  
    }  
    if (NTL_OVERFLOW(n, sizeof(T), 0))  
-      Error("excessive length in vector::SetLength"); 
+      ResourceError("excessive length in vector::SetLength"); 
       
    if (_vec__rep && NTL_VEC_HEAD(_vec__rep)->fixed) {
       if (NTL_VEC_HEAD(_vec__rep)->length == n) 
          return; 
       else 
-         Error("SetLength: can't change this vector's length"); 
+         LogicError("SetLength: can't change this vector's length"); 
    }  
 
    if (n == 0) {  
@@ -351,7 +466,7 @@ void Vec<T>::AllocateTo(long n)
       m = ((n+NTL_VectorMinAlloc-1)/NTL_VectorMinAlloc) * NTL_VectorMinAlloc; 
       char *p = (char *) NTL_SNS_MALLOC(m, sizeof(T), sizeof(_ntl_AlignedVectorHeader)); 
       if (!p) {  
-	 Error("out of memory in vector::SetLength()");  
+	 MemoryError();  
       }  
       _vec__rep = (T *) (p + sizeof(_ntl_AlignedVectorHeader)); 
   
@@ -363,13 +478,8 @@ void Vec<T>::AllocateTo(long n)
    else if (n > NTL_VEC_HEAD(_vec__rep)->alloc) {  
       m = max(n, long(NTL_VectorExpansionRatio*NTL_VEC_HEAD(_vec__rep)->alloc));  
       m = ((m+NTL_VectorMinAlloc-1)/NTL_VectorMinAlloc) * NTL_VectorMinAlloc; 
-      char *p = ((char *) _vec__rep) - sizeof(_ntl_AlignedVectorHeader); 
-      p = (char *) NTL_SNS_REALLOC(p, m, sizeof(T), sizeof(_ntl_AlignedVectorHeader)); 
-      if (!p) {  
-         Error("out of memory in vector::SetLength()");  
-      }  
-      _vec__rep = (T *) (p + sizeof(_ntl_AlignedVectorHeader)); 
-      NTL_VEC_HEAD(_vec__rep)->alloc = m;  
+
+      ReAllocate(m, IsRelocatableType((T*)0));
    }  
 }  
 
@@ -404,6 +514,25 @@ void Vec<T>::Init(long n, const T& src)
    AdjustMaxLength(n);
 }
 
+template<class T> template<class F>
+void Vec<T>::InitAndApply(long n, F& f)   
+{   
+   long num_init = MaxLength();
+   if (n <= num_init) return;
+
+   BlockConstruct(_vec__rep + num_init, n-num_init);
+
+   NTL_SCOPE(guard) { BlockDestroy(_vec__rep + num_init, n - num_init); };
+
+   long i;
+   for (i = num_init; i < n; i++)
+      f(_vec__rep[i]);
+
+   guard.relax();
+
+   AdjustMaxLength(n);
+}
+
 template<class T>
 void Vec<T>::DoSetLength(long n)
 {
@@ -419,13 +548,21 @@ void Vec<T>::DoSetLength(long n, const T& a)
    // a aliasing a vector element
    const T *src = &a;
    long pos = -1;
-   if (n >= allocated()) pos = position(a);
+   if (n > allocated()) pos = position(a);
    AllocateTo(n);
    if (pos != -1) src = elts() + pos;
    Init(n, *src);
    AdjustLength(n);
 }
 
+template<class T> template<class F>
+void Vec<T>::DoSetLengthAndApply(long n, F& f)
+{
+   AllocateTo(n);
+   InitAndApply(n, f);
+   AdjustLength(n);
+}
+
 
  
  
@@ -440,14 +577,17 @@ void Vec<T>::SetMaxLength(long n)
 template<class T>
 void Vec<T>::FixLength(long n) 
 { 
-   if (_vec__rep) Error("FixLength: can't fix this vector"); 
-   if (n < 0) Error("FixLength: negative length"); 
+   if (_vec__rep) LogicError("FixLength: can't fix this vector"); 
+   if (n < 0) LogicError("FixLength: negative length"); 
+
+   NTL_SCOPE(guard) { _vec__rep.kill(); };
+
    if (n > 0) 
       SetLength(n); 
    else { 
       char *p = (char *) NTL_SNS_MALLOC(0, sizeof(T), sizeof(_ntl_AlignedVectorHeader)); 
       if (!p) {  
-	 Error("out of memory in vector::FixLength()");  
+	 MemoryError();  
       }  
       _vec__rep = (T *) (p + sizeof(_ntl_AlignedVectorHeader)); 
   
@@ -456,14 +596,28 @@ void Vec<T>::FixLength(long n)
       NTL_VEC_HEAD(_vec__rep)->alloc = 0;  
    } 
    NTL_VEC_HEAD(_vec__rep)->fixed = 1; 
+
+   guard.relax();
 } 
+
+template<class T>
+void Vec<T>::FixAtCurrentLength() 
+{
+   if (fixed()) return;
+   if (length() != MaxLength()) 
+      LogicError("FixAtCurrentLength: can't fix this vector");
+
+   if (_vec__rep)
+      NTL_VEC_HEAD(_vec__rep)->fixed = 1;
+   else
+      FixLength(0);
+}
   
 template<class T>
 Vec<T>& Vec<T>::operator=(const Vec<T>& a)  
 {  
    if (this == &a) return *this;
 
-   long len = length();
    long init = MaxLength();
    long src_len = a.length();
    const T *src = a.elts();
@@ -494,44 +648,27 @@ template<class T>
 Vec<T>::~Vec()  
 {  
    if (!_vec__rep) return;  
-   BlockDestroy(_vec__rep, NTL_VEC_HEAD(_vec__rep)->init); 
-   NTL_SNS free(((char *) _vec__rep) - sizeof(_ntl_AlignedVectorHeader));  
+   BlockDestroy(_vec__rep.rep, NTL_VEC_HEAD(_vec__rep)->init); 
 }  
    
 template<class T>
 void Vec<T>::kill()  
 {  
-   if (!_vec__rep) return;  
-   if (NTL_VEC_HEAD(_vec__rep)->fixed) Error("can't kill this vector"); 
-   BlockDestroy(_vec__rep, NTL_VEC_HEAD(_vec__rep)->init); 
-   NTL_SNS free(((char *) _vec__rep) - sizeof(_ntl_AlignedVectorHeader));  
-   _vec__rep = 0; 
+   Vec<T> tmp;
+   this->swap(tmp);
 }  
   
-template<class T>
-void Vec<T>::RangeError(long i) const  
-{  
-   NTL_SNS cerr << "index out of range in vector: ";  
-   NTL_SNS cerr << i;  
-   if (!_vec__rep)  
-      NTL_SNS cerr << "(0)";  
-   else  
-      NTL_SNS cerr << "(" << NTL_VEC_HEAD(_vec__rep)->length << ")";  
-   Error("");  
-}  
   
 template<class T>
 void Vec<T>::swap(Vec<T>& y)  
 {  
-   T* t;  
    long xf = fixed();  
    long yf = y.fixed();  
    if (xf != yf ||   
        (xf && NTL_VEC_HEAD(_vec__rep)->length != NTL_VEC_HEAD(y._vec__rep)->length))  
-      Error("swap: can't swap these vectors");  
-   t = _vec__rep;  
-   _vec__rep = y._vec__rep;  
-   y._vec__rep = t;  
+      LogicError("swap: can't swap these vectors");  
+
+   _vec__rep.swap(y._vec__rep);
 } 
 
 template<class T>
@@ -540,6 +677,7 @@ void swap(Vec<T>& x, Vec<T>& y)
    x.swap(y);
 }
  
+// EXCEPTIONS: provides strong ES
 template<class T>
 void Vec<T>::append(const T& a)  
 {  
@@ -554,6 +692,9 @@ void Vec<T>::append(const T& a)
    if (len >= allocated()) pos = position(a);  
    AllocateTo(len+src_len);
 
+   // The logic here is copy-pasted from the append-vector
+   // logic...mostly
+
    long i;
    T *dst = elts();
    if (pos != -1) src = dst + pos;
@@ -565,7 +706,9 @@ void Vec<T>::append(const T& a)
    else {
       for (i = 0; i < init-len; i++)
          dst[i+len] = src[i];
-      Init(src_len+len, src+init-len);
+
+      // make sure we use BlockConstructFromObj
+      Init(src_len+len, *src);
    }
 
    AdjustLength(len+src_len);
@@ -617,7 +760,7 @@ NTL_SNS istream & operator>>(NTL_SNS istream& s, Vec<T>& a)
    Vec<T> ibuf;  
    long c;   
    long n;   
-   if (!s) Error("bad vector input"); 
+   if (!s) NTL_INPUT_ERROR(s, "bad vector input"); 
    
    c = s.peek();  
    while (IsWhiteSpace(c)) {  
@@ -625,7 +768,7 @@ NTL_SNS istream & operator>>(NTL_SNS istream& s, Vec<T>& a)
       c = s.peek();  
    }  
    if (c != '[') {  
-      Error("bad vector input");  
+      NTL_INPUT_ERROR(s, "bad vector input");  
    }  
    
    n = 0;   
@@ -641,14 +784,15 @@ NTL_SNS istream & operator>>(NTL_SNS istream& s, Vec<T>& a)
       if (n % NTL_VectorInputBlock == 0) ibuf.SetMaxLength(n + NTL_VectorInputBlock); 
       n++;   
       ibuf.SetLength(n);   
-      if (!(s >> ibuf[n-1])) Error("bad vector input");   
+      if (!(s >> ibuf[n-1])) NTL_INPUT_ERROR(s, "bad vector input");   
       c = s.peek();  
       while (IsWhiteSpace(c)) {  
          s.get();  
          c = s.peek();  
       }  
    }   
-   if (IsEOFChar(c)) Error("bad vector input");  
+
+   if (IsEOFChar(c)) NTL_INPUT_ERROR(s, "bad vector input");  
    s.get(); 
    
    a = ibuf; 
@@ -703,6 +847,7 @@ void conv(Vec<T>& x, const Vec<S>& a)
 {
    long n = a.length();
    x.SetLength(n);
+
    for (long i = 0; i < n; i++)
       conv(x[i], a[i]);
 }
